package blackbox;

import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.StringReader;
import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Scanner;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JTabbedPane;
import javax.swing.KeyStroke;
import javax.swing.TransferHandler;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

import blackbox.Output.Mapping;

/**
 * Main BlackBox3 class
 *
 * @author Robbie McElrath
 */
public class BlackBox3 extends JFrame implements ActionListener, ChangeListener, WindowListener {
  private final static long serialVersionUID = 387983748234L;

  /** Filename for BlackBox2/3 settings file */
  private final static String SETTINGS_NAME = "BB3Settings.xml";
  /** Prefix for the window's title */
  private final static String WINDOW_TITLE = "BlackBox3 - ";
  /** BlackBox Icon */
  private final static Image icon = Toolkit.getDefaultToolkit().getImage(
      BlackBox3.class.getResource("res/icon.gif"));
  /** URL of online help */
  private final static String HELP_URL = "http://www.robbiemcelrath.com/fs/blackbox/help";
  /** Prefix for all BlackBox2/3 generated gauges */
  final static String GAUGE_HEADER = "<!-- BlackBox2 Gauge";
  /** BlackBox version number */
  final static String VERSION = "3.1.0";

  /** The BlackBox gauge that was last opened */
  private File lastOpened;
  /** BBXML filter used to ensure files are valid */
  private BBXMLFilter bbXMLFilter;
  /** The JTabbedPane that stores all the tabs */
  private JTabbedPane tabbedPane;
  /** Whether or not the settings (output formats, last opened file) have been changed. */
  private boolean settingsChanged = false;
  /** Used to generate the BlackBox gauges */
  private Output output = new Output();

  /** Column output format */
  private JRadioButtonMenuItem outColumn;
  /** Wrapped column output format */
  private JRadioButtonMenuItem outWrapped;
  /** Tabbed output format */
  private JRadioButtonMenuItem outTabbed;
  /** Multi-Tabbed output format */
  private JRadioButtonMenuItem outMultiTabbed;
  /** Logger CSV output format */
  private JRadioButtonMenuItem outLoggerCsv;
  /** Logger line output format */
  private JRadioButtonMenuItem outLoggerLine;

  /** Check box for including units */
  JCheckBoxMenuItem showUnitsCheckBox;

  /* The JRadioButtonMenuItems used for sorting */
  private JRadioButtonMenuItem sOccurence;
  private JRadioButtonMenuItem sType;
  private JRadioButtonMenuItem sName;
  private JRadioButtonMenuItem sUnits;
  private JRadioButtonMenuItem sChecked;


  /**
   * Default constructor
   */
  public BlackBox3() {
    super(WINDOW_TITLE + "New");
    setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    setTransferHandler(new TransHandler());
    setMinimumSize(new Dimension(150, 150));
    setLayout(new GridBagLayout());
    GridBagConstraints c = createConstraints();
    addWindowListener(this);
    bbXMLFilter = new BBXMLFilter();
    // setup the GUI
    try {
      UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
      setIconImage(icon);
    } catch (Throwable e) {
    }
    // instantiate some of the instance variables
    lastOpened = new File(System.getProperty("user.dir"));
    output.setBlackBox3(this);
    setJMenuBar(MenuBar.createMenuBar()
        .addMenu(MenuBar.createMenu("File")
            .addItem(newGauge)
            .addItem(open)
            .addItem(save)
            .addItem(saveAs)
            .addSeperator()
            .addItem(exit))
        .addMenu(MenuBar.createMenu("Edit")
            .addItem(add)
            .addSeperator()
            .addItem(reparse)
            .addItem(selectAll)
            .addItem(selectNone)
            .addMenu(MenuBar.createMenu("Sort By...", getIcon("sort"))
                .addRadioGroup(MenuBar.createRadioGroup()
                    .addRadioItem(sOccurence = new JRadioButtonMenuItem(sortOccurence))
                    .addRadioItem(sType = new JRadioButtonMenuItem(sortType))
                    .addRadioItem(sName = new JRadioButtonMenuItem(sortName))
                    .addRadioItem(sUnits = new JRadioButtonMenuItem(sortUnits))
                    .addRadioItem(sChecked = new JRadioButtonMenuItem(sortChecked))))
            .addSeperator()
            .addCheckItem(showUnitsCheckBox = new JCheckBoxMenuItem(showUnits))
            .addMenu(MenuBar.createMenu("Output Format...", getIcon("output"))
                .addRadioGroup(MenuBar.createRadioGroup()
                    .addRadioItem(outColumn = new JRadioButtonMenuItem(output.column))
                    .addRadioItem(outWrapped = new JRadioButtonMenuItem(output.wrapped))
                    .addRadioItem(outTabbed = new JRadioButtonMenuItem(output.tabbed))
                    .addRadioItem(outMultiTabbed = new JRadioButtonMenuItem(output.multiTabbed))
                    .addRadioItem(outLoggerCsv = new JRadioButtonMenuItem(output.loggerCsv))
                    .addRadioItem(outLoggerLine = new JRadioButtonMenuItem(output.loggerLine)))))
        .addMenu(MenuBar.createMenu("Help").addItem(help).addItem(about)).getJMenuBar());
    add(ToolBar.createToolBar()
        .addItem(add)
        .addItem(save)
        .addSeperator()
        .addItem(reparse)
        .addSeperator()
        .addItem(selectAll)
        .addItem(selectNone)
        .getJToolBar(), c);

    // set default output format
    outColumn.setSelected(true);
    // add the tabbedpane
    c.gridy++;
    c.weightx = c.weighty = 1.0;
    tabbedPane = new DnDTabbedPane();
    tabbedPane.addChangeListener(this);
    add(tabbedPane, c);
    // disable all inactive actions
    setActionsEnabled(false);
    // read the settings XML file
    InputStream settings;
    try {
      settings = new FileInputStream(SETTINGS_NAME);
    } catch (FileNotFoundException e) {
      settings = BlackBox3.class.getResourceAsStream("res/settings.xml");
    }
    try {
      XMLReader reader = XMLReaderFactory.createXMLReader();
      reader.setContentHandler(new SettingsHandler());
      reader.parse(new InputSource(settings));
      settings.close();
    } catch (IOException e) {
      showError("Could not read the settings file!");
    } catch (SAXException e) {
      showError("An error occured while parsing the settings file!");
    }
    // open the last saved XML file
    if (lastOpened.isFile()) {
      open(lastOpened);
    }
  }


  // --------------------------------------------------------------------------------------- ACTIONS
  /**
   * Action for creating a new BlackBox gauge
   */
  private Action newGauge = new BB2Action("New", getIcon("new"),
      "Creates a new, empty BlackBox gauge.", KeyEvent.VK_N, KeyStroke.getKeyStroke("ctrl N")) {
    private final static long serialVersionUID = 5482934342L;

    @Override
    public void actionPerformed(ActionEvent event) {
      // check if everything's saved or not
      if (!promptToSave(true)) {
        return;
      }
      // everything's saved, so close the gauge and start a new one
      tabbedPane.removeAll();
      lastOpened = null;
      setActionsEnabled(false);
      setTitle(WINDOW_TITLE + "New");
    }
  };

  /**
   * Action for opening a BlackBox gauge
   */
  private Action open = new BB2Action("Open", getIcon("open"), "Opens an existing BlackBox gauge.",
      KeyEvent.VK_O, KeyStroke.getKeyStroke("ctrl O")) {
    private final static long serialVersionUID = 598787342334L;

    @Override
    public void actionPerformed(ActionEvent event) {
      JFileChooser fc = new JFileChooser(lastOpened);
      fc.setFileFilter(bbXMLFilter);
      if (fc.showOpenDialog(BlackBox3.this) == JFileChooser.APPROVE_OPTION) {
        open(fc.getSelectedFile());
      }
    }
  };

  /**
   * Action for prompting for adding a new gauge
   */
  private Action add = new BB2Action("Add Gauge", getIcon("add"), "Adds a gauge in a new tab.",
      KeyEvent.VK_G, KeyStroke.getKeyStroke("ctrl G")) {
    private final static long serialVersionUID = 598787342334L;

    @Override
    public void actionPerformed(ActionEvent event) {
      File dir = getSelectedGauge() == null ? lastOpened : getSelectedGauge().getFile();
      JFileChooser fc = new JFileChooser(dir);
      fc.setFileFilter(new XMLFilter());
      if (fc.showOpenDialog(BlackBox3.this) == JFileChooser.APPROVE_OPTION) {
        addGauge(fc.getSelectedFile());
      }
    }
  };

  /**
   * Adds the selected file as a new gauge in a new tab
   *
   * @param gaugeLoc The File representing the Gauge to import
   */
  private Gauge addGauge(File gaugeLoc) {
    return addGauge(gaugeLoc, null);
  }

  /**
   * Adds the selected file as a new gauge and pre-selects the variables in the List.
   *
   * @param gaugeLoc The File representing the Gauge to import
   * @param vars The variables to select by default (in the format Variable.toString() puts them in)
   */
  private Gauge addGauge(File gaugeLoc, ArrayList<String> vars) {
    Gauge g = new Gauge(gaugeLoc, vars);
    if (hasGauge(g)) {
      tabbedPane.setSelectedComponent(g);
    } else {
      tabbedPane.addTab(g.getName(), g);
      tabbedPane.setTabComponentAt(tabbedPane.getTabCount() - 1, g.getTab());
      tabbedPane.setSelectedIndex(tabbedPane.getTabCount() - 1);
      setActionsEnabled(true);
    }
    gaugeChanged();
    g.setSaved();
    return g;
  }

  /**
   * Action for prompting for saving
   */
  private Action save = new BB2Action("Save", getIcon("save"),
      "Saves the selected variables into the BlackBox gauge.", KeyEvent.VK_S,
      KeyStroke.getKeyStroke("ctrl S")) {
    private final static long serialVersionUID = 598787342334L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (lastOpened == null || !lastOpened.isFile()) {
        saveAs.actionPerformed(event);
      } else {
        // last saved file exists, make sure we can overwrite it
        if (bbXMLFilter.accept(lastOpened)) {
          save(lastOpened);
        } else {
          saveAs.actionPerformed(event);
        }
      }
    }
  };

  /**
   * Action for prompting for 'save as'
   */
  private Action saveAs = new BB2Action("Save As...", getIcon("saveAs"),
      "Saves the selected variables into a new BlackBox gauge.", KeyEvent.VK_A,
      KeyStroke.getKeyStroke("shift ctrl S")) {
    private final static long serialVersionUID = 598787342334L;

    @Override
    public void actionPerformed(ActionEvent event) {
      File dir = getSelectedGauge() == null ? lastOpened : getSelectedGauge().getFile();
      JFileChooser fc = new JFileChooser(dir);
      fc.setFileFilter(bbXMLFilter);
      if (fc.showSaveDialog(BlackBox3.this) == JFileChooser.APPROVE_OPTION) {
        // they made a selection, which should be valid, but double check...
        File file = fc.getSelectedFile();
        if (!file.getName().toLowerCase().endsWith(".xml")) {
          file = new File(file.getAbsolutePath() + ".xml");
        }
        if (file.exists()) {
          if (JOptionPane.showConfirmDialog(BlackBox3.this, "The file \"" + file.getName()
              + "\" already exists! " + "Do you want to overwrite it?", "BlackBox3",
              JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE) != JOptionPane.YES_OPTION) {
            return;
          }
          if (!bbXMLFilter.accept(file) && JOptionPane.showConfirmDialog(
              BlackBox3.this, "\"" + file.getName() + "\" was not created by BlackBox! You may be "
                  + "overwriting your own gauge!\nAre you sure you wish to continue?", "BlackBox3",
              JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE) != JOptionPane.YES_OPTION) {
            return;
          }
        }
        save(file);
      }
    }
  };

  /**
   * Action for showing/hiding units
   */
  private Action showUnits = new BB2Action("Show Units", null,
      "Toggles whether units should be included on the variables in the generated BlackBox gauge.",
      KeyEvent.VK_U, null) {
    private final static long serialVersionUID = 598787342334L;

    @Override
    public void actionPerformed(ActionEvent event) {
      setModified();
    }
  };

  /**
   * Action for reloading a gauge
   */
  private Action reparse = new BB2Action("Reparse", getIcon("reparse"),
      "Re-parses the currently selected file.", KeyEvent.VK_R, KeyStroke.getKeyStroke("F5")) {
    private final static long serialVersionUID = 598787342334L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (hasGauge()) {
        getSelectedGauge().reload();
        gaugeChanged();
      }
    }
  };

  /**
   * Action for selecting all the variables
   */
  private Action selectAll = new BB2Action("Select All", null,
      "Selects all the variable in this gauge.", KeyEvent.VK_A, KeyStroke.getKeyStroke("ctrl A")) {
    private final static long serialVersionUID = 598787342334L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (hasGauge()) {
        int result =
            JOptionPane.showConfirmDialog(BlackBox3.this,
                "Are you sure you want to select all variables?", "BlackBox3",
                JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
        if (result == JOptionPane.YES_OPTION) {
          getSelectedGauge().selectAll();
        }
      }
    }
  };

  /**
   * Action for selecting none the variables
   */
  private Action selectNone =
      new BB2Action("Select None", null, "Deselects all the variable in this gauge.",
          KeyEvent.VK_D, KeyStroke.getKeyStroke("ctrl D")) {
        private final static long serialVersionUID = 598787342334L;

        @Override
        public void actionPerformed(ActionEvent event) {
          if (hasGauge()) {
            int result =
                JOptionPane.showConfirmDialog(BlackBox3.this,
                    "Are you sure you want to clear all variables?", "BlackBox3",
                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
            if (result == JOptionPane.YES_OPTION) {
              getSelectedGauge().deselectAll();
            }
          }
        }
      };

  /**
   * Default action for exiting the program
   */
  private Action exit = new BB2Action("Exit", null, "Quits BlackBox", KeyEvent.VK_X,
      KeyStroke.getKeyStroke("alt F4")) {
    private final static long serialVersionUID = 54987238435L;

    @Override
    public void actionPerformed(ActionEvent event) {
      saveSettings();
      System.exit(0);
    }
  };

  /**
   * Action for sorting by occurrence
   */
  private Action sortOccurence = new BB2Action("Occurence", null,
      "Sorts the variables by their occurence in the source file.", null, null) {
    private final static long serialVersionUID = 54987238435L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (hasGauge() && getSelectedGauge().getSortedBy() != Gauge.SortBy.OCCURRENCE) {
        getSelectedGauge().sort(Gauge.SortBy.OCCURRENCE);
      }
    }
  };

  /**
   * Action for sorting by type
   */
  private Action sortType = new BB2Action("Type", null, "Sorts the variables by their type.", null,
      null) {
    private final static long serialVersionUID = 54987238435L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (hasGauge() && getSelectedGauge().getSortedBy() != Gauge.SortBy.TYPE) {
        getSelectedGauge().sort(Gauge.SortBy.TYPE);
      }
    }
  };

  /**
   * Action for sorting by name
   */
  private Action sortName = new BB2Action("Name", null, "Sorts the variables by their name.", null,
      null) {
    private final static long serialVersionUID = 54987238435L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (hasGauge() && getSelectedGauge().getSortedBy() != Gauge.SortBy.NAME) {
        getSelectedGauge().sort(Gauge.SortBy.NAME);
      }
    }
  };

  /**
   * Action for sorting by units
   */
  private Action sortUnits = new BB2Action("Units", null, "Sorts the variables by their units.",
      null, null) {
    private final static long serialVersionUID = 54987238435L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (hasGauge() && getSelectedGauge().getSortedBy() != Gauge.SortBy.UNITS) {
        getSelectedGauge().sort(Gauge.SortBy.UNITS);
      }
    }
  };

  /**
   * Action for sorting by checked status
   */
  private Action sortChecked = new BB2Action("Checked", null,
      "Sorts the variables by their checked status.", null, null) {
    private final static long serialVersionUID = 54987238435L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (hasGauge() && getSelectedGauge().getSortedBy() != Gauge.SortBy.CHECKED) {
        getSelectedGauge().sort(Gauge.SortBy.CHECKED);
      }
    }
  };

  /**
   * Help action
   */
  private Action help = new BB2Action("Online Help", null, "Opens online help.", null,
      KeyStroke.getKeyStroke("F1")) {
    private final static long serialVersionUID = -98787234L;

    @Override
    public void actionPerformed(ActionEvent event) {
      if (Desktop.isDesktopSupported()) {
        try {
          Desktop.getDesktop().browse(new URI(HELP_URL));
        } catch (Exception e) {
          showError("Error opening help!\nPlease visit " + HELP_URL + ".");
        }
      } else {
        JOptionPane.showMessageDialog(BlackBox3.this, "Please visit " + HELP_URL
            + " for online help.", "BlackBox3", JOptionPane.INFORMATION_MESSAGE);
      }
    }
  };

  /**
   * About dialog action
   */
  private Action about = new BB2Action("About", null, "About BlackBox3", null, null) {
    private final static long serialVersionUID = 598798234L;

    @Override
    public void actionPerformed(ActionEvent event) {
      Dialogs.about(BlackBox3.this);
    }
  };

  /**
   * A list of all actions that depend on a gauge being currently opened. These actions will be
   * disabled when no gauge is open
   */
  private Action[] dependent = new Action[] {save, saveAs, reparse, selectAll, selectNone,
      sortOccurence, sortType, sortName, sortUnits, sortChecked, output.column, output.wrapped,
      output.tabbed, output.multiTabbed, output.loggerCsv, output.loggerLine};

  /**
   * Action template
   */
  protected abstract class BB2Action extends AbstractAction {
    private final static long serialVersionUID = 5494893894523L;

    /**
     * Creates a new Action with the specified properties. Leave any of the parameters null if you
     * don't wish to use them
     *
     * @param text The name of the action (will appear on a button or menu item)
     * @param icon The icon to display with this event
     * @param desc A short description of the event for use in a tooltip
     * @param mnemonic The KeyEvent.VK_* value for use with ALT+x combinations
     * @param accelerator The KeyStroke shortcut for the action
     */
    public BB2Action(String text, Icon icon, String desc, Integer mnemonic, KeyStroke accelerator) {
      super(text, icon);
      putValue(AbstractAction.SHORT_DESCRIPTION, desc);
      putValue(AbstractAction.MNEMONIC_KEY, mnemonic);
      putValue(AbstractAction.ACCELERATOR_KEY, accelerator);
    }

    /**
     * Returns the name of the action
     */
    @Override
    public String toString() {
      return getValue(Action.NAME).toString();
    }
  }


  // ----------------------------------------------------------------------------- OPEN/SAVE METHODS
  /**
   * Opens the specified BlackBox gauge
   *
   * @param file The gauge to open. It should be a BlackBox generated gauge
   */
  private void open(File file) {
    try {
      Scanner xml = new Scanner(file);
      if (xml.hasNextLine() && xml.nextLine().equals(GAUGE_HEADER)) {
        StringBuilder comment = new StringBuilder();
        while (xml.hasNextLine()) {
          String line = xml.nextLine();
          if (line.trim().equals("-->")) {
            break;
          }
          comment.append(line + "\r\n");
        }
        String info = comment.toString();
        XMLReader reader = XMLReaderFactory.createXMLReader();
        reader.setContentHandler(new GaugeHandler());
        reader.parse(new InputSource(new StringReader(info)));
      } else {
        // it's not a valid BlackBox gauge
        showError("\"" + file.getName() + "\" is not a valid BlackBox gauge!\nIf you "
            + "would like to add an XML gauge to the BlackBox gauge you're creating, "
            + "press the \"Add Gauge\" button in the toolbar.");
      }
      xml.close();
      setTitle(WINDOW_TITLE + file.getName());
      if (lastOpened != null && !lastOpened.getAbsolutePath().equals(file.getAbsolutePath())) {
        settingsChanged = true;
      }
      lastOpened = file;
    } catch (FileNotFoundException e) {
      showError("Could not open \"" + file.getName() + "\"!");
    } catch (IOException e) {
      showError("An I/O Error occured while opening \"" + file.getName() + "\"!");
    } catch (SAXException e) {
      showError("An error occured while parsing \"" + file.getName() + "\"'s information!");
    }
  }

  /**
   * Performs the saving of the gauge as the specified file. If the file already exits, it will be
   * overwritten - as such, all overwrite checking should be done prior to calling this method.
   *
   * @param file The file to output the gauge to
   */
  private void save(File file) {
    try {
      PrintStream out = new PrintStream(file);
      // output BlackBox info
      out.println(GAUGE_HEADER);
      out.println("\t<blackbox>");
      out.println("\t\t<output format=\"" + output.format.getName() + "\" />");
      out.println("\t\t<display units=\"" + (showUnitsCheckBox.isSelected() ? "yes" : "no")
          + "\" />");
      GaugeIterator gauges = new GaugeIterator();
      while (gauges.hasNext()) {
        Gauge gauge = gauges.next();
        out.println("\t\t<gauge file=\"" + gauge.getFile().getPath() + "\">");
        Iterator<Variable> vars = gauge.getSelected();
        while (vars.hasNext()) {
          out.println("\t\t\t<variable name=\"" + vars.next().toString() + "\" />");
        }
        out.println("\t\t</gauge>");
      }
      out.println("\t</blackbox>");
      out.println("-->");
      // now output everything
      if (output.format != null) {
        output.format.print(out, gauges.reset());
      }
      out.close();
      setTitle(WINDOW_TITLE + file.getName());
      if (!lastOpened.getAbsolutePath().equals(file.getAbsolutePath())) {
        settingsChanged = true;
      }
      lastOpened = file;
      // set the tabs as saved
      gauges.reset();
      while (gauges.hasNext()) {
        gauges.next().setSaved();
      }
    } catch (FileNotFoundException e) {
      String[] options = {"Try Again", "Cancel"};
      if (JOptionPane.showOptionDialog(this, "It appears Flight Simulator is currently using this "
          + "file!\nPlease unload either the aircraft or the gauge, then try saving again.",
          "BlackBox3", JOptionPane.OK_CANCEL_OPTION, JOptionPane.ERROR_MESSAGE, null, options,
          options[1]) == JOptionPane.OK_OPTION) {
        save(file);
      }
    }
  }


  // ---------------------------------------------------------------------------------- XML HANDLERS
  /**
   * Handles the Settings XML file
   */
  private class SettingsHandler extends DefaultHandler {
    /** Points to the TreeMap we're currently filling */
    private ArrayList<Output.Mapping> current;

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attribs) {
      if (localName.equals("file")) {
        String file = attribs.getValue("name");
        if (file != null && file.trim() != "") {
          lastOpened = new File(file);
        }
      } else if (localName.equals("gauge")) {
        String height = attribs.getValue("height");
        String width = attribs.getValue("width");
        String bgcolor = attribs.getValue("fillcolor");
        String font = attribs.getValue("font");
        if (bgcolor != null) {
          output.backgroundColor = bgcolor;
        }
        if (font != null) {
          output.font = font;
        }
        try {
          if (height != null) {
            output.height = Integer.parseInt(height);
          }
          if (width != null) {
            output.width = Integer.parseInt(width);
          }
        } catch (NumberFormatException e) {
        }
      } else if (localName.equals("title")) {
        String color = attribs.getValue("color");
        if (color != null) {
          output.titleColor = color;
        }
      } else if (localName.equals("variable")) {
        String color = attribs.getValue("color");
        String format = attribs.getValue("format");
        if (color != null) {
          output.variableColor = color;
        }
        if (format != null) {
          output.mappingDefault = new Output.Mapping("", format);
        }
      } else if (localName.equals("tab")) {
        String color = attribs.getValue("color");
        String background = attribs.getValue("background");
        String selectedColor = attribs.getValue("selectedColor");
        String selectedBackground = attribs.getValue("selectedBackground");
        if (color != null) {
          output.tabColor = color;
        }
        if (background != null) {
          output.tabBackground = background;
        }
        if (selectedColor != null) {
          output.tabSelectedColor = selectedColor;
        }
        if (selectedBackground != null) {
          output.tabSelectedBackground = selectedBackground;
        }
      } else if (localName.equals("logger")) {
        String path = attribs.getValue("path");
        if (path != null) {
          output.loggerPath = path;
        }
      } else if (localName.equals("units")) {
        current = output.mappingUnits = new ArrayList<Output.Mapping>();
      } else if (localName.equals("name")) {
        current = output.mappingName = new ArrayList<Output.Mapping>();
      } else if (localName.equals("equate")) {
        String key = attribs.getValue("key");
        String value = attribs.getValue("value");
        if (key != null && value != null && current != null) {
          current.add(new Output.Mapping(key, value));
        }
      }
    }
  }

  /**
   * Parses the gauge information placed in a comment of every BlackBox generated gauge.
   */
  private class GaugeHandler extends DefaultHandler {
    private String gaugeLoc;
    private ArrayList<String> vars;

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attribs) {
      if (localName.equals("gauge")) {
        gaugeLoc = attribs.getValue("file");
        if (gaugeLoc != null) {
          vars = new ArrayList<String>();
        } else {
          vars = null;
        }
      } else if (localName.equals("variable")) {
        String var = attribs.getValue("name");
        if (var != null && vars != null) {
          vars.add(var);
        }
      } else if (localName.equals("output")) {
        String format = attribs.getValue("format");
        if (format.equals(output.column.getName())) {
          outColumn.setSelected(true);
          output.format = output.column;
        } else if (format.equals(output.wrapped.getName())) {
          outWrapped.setSelected(true);
          output.format = output.wrapped;
        } else if (format.equals(output.tabbed.getName())) {
          outTabbed.setSelected(true);
          output.format = output.tabbed;
        } else if (format.equals(output.multiTabbed.getName())) {
          outMultiTabbed.setSelected(true);
          output.format = output.multiTabbed;
        } else if (format.equals(output.loggerCsv.getName())) {
          outLoggerCsv.setSelected(true);
          output.format = output.loggerCsv;
        } else if (format.equals(output.loggerLine.getName())) {
          outLoggerLine.setSelected(true);
          output.format = output.loggerLine;
        }
      } else if (localName.equals("display")) {
        String units = attribs.getValue("units");
        showUnitsCheckBox.setSelected(units != null && units.equals("yes"));
      }
    }

    @Override
    public void endElement(String uri, String localName, String qName) {
      if (localName.equals("gauge")) {
        // have gauge location & its selected variables, now add it
        if (gaugeLoc != null && vars != null) {
          File gauge = new File(gaugeLoc);
          if (gauge.isFile() && gauge.canRead()) {
            addGauge(gauge, vars);
          }
        }
      }
    }
  }


  // -------------------------------------------------------------------------------- GAUGE ITERATOR
  class GaugeIterator implements Iterator<Gauge> {
    int index = 0;

    public GaugeIterator() {
      index = 0;
    }

    @Override
    public boolean hasNext() {
      return index < tabbedPane.getTabCount();
    }

    @Override
    public Gauge next() {
      return (Gauge) tabbedPane.getComponentAt(index++);
    }

    public GaugeIterator reset() {
      index = 0;
      return this;
    }

    public int getIndex() {
      return index - 1;
    }

    public int getSize() {
      return tabbedPane.getTabCount();
    }

    public GaugeIterator createNew() {
      return new GaugeIterator();
    }

    @Override
    public void remove() {}
  }


  // ----------------------------------------------------------------------------------- DRAG & DROP
  /**
   * This class handles Drag and Drop behavior onto the BlackBox window.
   *
   * @author Robbie McElrath
   */
  public class TransHandler extends TransferHandler {
    private final static long serialVersionUID = 5908709283343L;

    @Override
    public boolean canImport(TransferHandler.TransferSupport support) {
      return support.isDataFlavorSupported(DataFlavor.javaFileListFlavor);
    }

    @Override
    @SuppressWarnings("unchecked")
    public boolean importData(TransferHandler.TransferSupport support) {
      if (!canImport(support)) {
        return false;
      }
      Transferable t = support.getTransferable();
      try {
        List<File> files = (List<File>) t.getTransferData(DataFlavor.javaFileListFlavor);
        for (File f : files) {
          addGauge(f);
        }
      } catch (UnsupportedFlavorException e) {
        return false;
      } catch (IOException e) {
        return false;
      } catch (ClassCastException e) {
        return false;
      }
      return true;
    }
  }


  // -------------------------------------------------------------------------- HELPER/OTHER METHODS
  /* Unused listener methods */
  @Override
  public void actionPerformed(ActionEvent e) {}

  @Override
  public void windowActivated(WindowEvent e) {}

  @Override
  public void windowClosed(WindowEvent e) {}

  @Override
  public void windowDeactivated(WindowEvent e) {}

  @Override
  public void windowDeiconified(WindowEvent e) {}

  @Override
  public void windowIconified(WindowEvent e) {}

  @Override
  public void windowOpened(WindowEvent e) {}

  @Override
  public void windowClosing(WindowEvent e) {
    promptToSave(false);
    saveSettings();
  }

  /**
   * Prompts the user if they want to save if it isn't already
   *
   * @return true if the user clicked yes or no, false if they said cancel
   */
  private boolean promptToSave(boolean cancel) {
    boolean saved = true;
    Iterator<Gauge> gauges = new GaugeIterator();
    while (gauges.hasNext()) {
      saved &= gauges.next().isSaved();
    }
    if (!saved) {
      int response =
          JOptionPane.showConfirmDialog(BlackBox3.this, "Would you like to save "
              + "the gauge you're working on?", "BlackBox3",
              cancel ? JOptionPane.YES_NO_CANCEL_OPTION : JOptionPane.YES_NO_OPTION,
              JOptionPane.QUESTION_MESSAGE);
      if (response == JOptionPane.YES_OPTION) {
        save.actionPerformed(null);
      } else if (response == JOptionPane.CANCEL_OPTION) {
        return false;
      }
    }
    return true;
  }

  /**
   * Saves all settings prior to quitting.
   */
  private void saveSettings() {
    if (settingsChanged) {
      // save settings
      try (PrintStream settings = new PrintStream(SETTINGS_NAME)) {
        settings.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
        settings.println("<settings>");
        if (lastOpened.isFile()) {
          settings.println("\t<file name=\"" + lastOpened.getPath() + "\" />");
        }
        settings.println("\t<gauge width=\"" + output.width + "\" height=\"" + output.height
            + "\" fillcolor=\"" + output.backgroundColor + "\" font=\"" + output.font + "\" />");
        settings.println("\t<title color=\"" + output.titleColor + "\" />");
        settings.println("\t<variable color=\"" + output.variableColor + "\" format=\""
            + output.mappingDefault.getFormat() + "\" />");
        settings.println("\t<tab color=\"" + output.tabColor + "\" background=\""
            + output.tabBackground + "\" selectedColor=\"" + output.tabSelectedColor
            + "\" selectedBackground=\"" + output.tabSelectedBackground + "\" />");
        settings.println("\t<logger path=\"" + output.loggerPath + "\" />");
        settings.println("\t<!-- ALL keys should be in lower case -->");
        settings.println("\t<units>");
        for (Mapping map : output.mappingUnits) {
          settings.println("\t\t<equate key=\"" + map.getKey() + "\" value=\"" + map.getFormat()
              + "\" />");
        }
        settings.println("\t</units>");
        settings.println("\t<name>");
        for (Mapping map : output.mappingName) {
          settings.println("\t\t<equate key=\"" + map.getKey() + "\" value=\"" + map.getFormat()
              + "\" />");
        }
        settings.println("\t</name>");
        settings.println("</settings>");
      } catch (FileNotFoundException err) {
        showError("BlackBox could not create the settings file!");
      }
    }
  }

  /**
   * Enables/disables all of the dependent actions (those that require a gauge to be open to
   * function.
   *
   * @param enable True to enable the dependent actions
   */
  private void setActionsEnabled(boolean enable) {
    for (Action actn : dependent) {
      actn.setEnabled(enable);
    }

  }

  /**
   * Checks if there is a gauge open
   *
   * @return True if at least one gauge is open
   */
  private boolean hasGauge() {
    return tabbedPane != null && tabbedPane.getSelectedComponent() != null;
  }

  /**
   * Checks if the given gauge is opened
   *
   * @param gauge The gauge to check for
   * @return true if the gauge is open
   */
  private boolean hasGauge(Gauge gauge) {
    Iterator<Gauge> gauges = new GaugeIterator();
    while (gauges.hasNext()) {
      if (gauges.next().equals(gauge)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Returns the currently selected gauge
   *
   * @return The gauge that's open
   */
  private Gauge getSelectedGauge() {
    return (Gauge) tabbedPane.getSelectedComponent();
  }

  /**
   * Closes the given Gauge if it is open
   *
   * @param gauge The gauge to close
   * @return true if the gauge was closed, false if it didn't exist
   */
  boolean close(Gauge gauge) {
    if (!hasGauge(gauge)) {
      return false;
    }
    // confirm that they want to close it
    if (JOptionPane.showConfirmDialog(this, "Are you sure you want to remove this gauge?",
        "BlackBox3", JOptionPane.YES_NO_OPTION,
        JOptionPane.QUESTION_MESSAGE) == JOptionPane.YES_OPTION) {
      if (gauge.hasSelected()) {
        setModified();
      }
      tabbedPane.remove(gauge);
      if (tabbedPane.getTabCount() == 0) {
        setActionsEnabled(false);
      }
      return true;
    }
    return false;
  }

  /**
   * Sets the BlackBox gauge as modified
   */
  void setModified() {
    if (getTitle().charAt(WINDOW_TITLE.length()) != '*') {
      setTitle(WINDOW_TITLE + '*' + getTitle().substring(WINDOW_TITLE.length()));
    }
  }

  /**
   * Called when the selected tab changes. It updates the title of the window and the sorting menu
   * to account for the changes.
   *
   * @param event The ChangeEvent object
   */
  @Override
  public void stateChanged(ChangeEvent event) {
    gaugeChanged();
  }

  /**
   * Updates the menu items to account for the selected gauge being changed
   */
  void gaugeChanged() {
    Gauge.SortBy sb = null;
    if (hasGauge()) {
      sb = getSelectedGauge().getSortedBy();
    }
    sOccurence.setSelected(sb == Gauge.SortBy.OCCURRENCE);
    sType.setSelected(sb == Gauge.SortBy.TYPE);
    sName.setSelected(sb == Gauge.SortBy.NAME);
    sUnits.setSelected(sb == Gauge.SortBy.UNITS);
    sChecked.setSelected(sb == Gauge.SortBy.CHECKED);

  }

  /**
   * Displays a message dialog with the given body text, an error icon, and a title of "Error!"
   *
   * @param text The text of the error message
   */
  private void showError(String text) {
    JOptionPane.showMessageDialog(this, text, "BlackBox3", JOptionPane.ERROR_MESSAGE);
  }

  /**
   * Generates a child window with the given size centered around the position of the parent
   * BlackBox3 window. Only size, location, and title are set by this method. All other settings are
   * left as default and must be changed by the client code.
   *
   * @param name The title of the window
   * @param height The height in pixels of the window
   * @param width The width in pixels of the window
   * @return A new JFrame with the specified attributes
   */
  JFrame createWindow(String name, int height, int width) {
    JFrame win = new JFrame(name);
    win.setSize(width, height);
    win.setIconImage(icon);
    Point p = getCenter();
    p.x -= width / 2;
    p.y -= height / 2;
    win.setLocation(p);
    return win;
  }

  /**
   * Returns the location of the center of the BlackBox3 window relative to the upper-left corner of
   * the screen
   *
   * @return The locatin of the center of the screen
   */
  private Point getCenter() {
    Dimension d = getSize();
    Point p = getLocation();
    p.x += d.width / 2;
    p.y += d.height / 2;
    return p;
  }


  // -------------------------------------------------------------------------------- STATIC METHODS
  /**
   * Entry point
   *
   * @param args
   */
  public static void main(String[] args) {
    BlackBox3 win = new BlackBox3();
    win.setSize(500, 500);
    Dimension scrn = Toolkit.getDefaultToolkit().getScreenSize();
    win.setLocation((scrn.width - win.getWidth()) / 2, (scrn.height - win.getHeight()) / 2);
    win.setVisible(true);
  }

  /**
   * Returns the Icon with the given name
   *
   * @param name The filename of the icon (relative to the 'res' directory, without extension)
   * @return The Icon
   */
  static Icon getIcon(String name) {
    try {
      return new ImageIcon(BlackBox3.class.getResource("res/" + name + ".png"));
    } catch (Throwable t) {
      return null;
    }
  }

  /**
   * Creates a new GridBagConstraints object initialized with the normal values.
   *
   * @return A new GridBagConstraint
   */
  public static GridBagConstraints createConstraints() {
    GridBagConstraints c = new GridBagConstraints();
    c.gridx = 0;
    c.gridy = 0;
    c.fill = GridBagConstraints.BOTH;
    return c;
  }

  /**
   * Replaces the first occurrence of a substring with another
   *
   * @param main The string in which to replace text
   * @param target The string that will be replaced
   * @param replacement The string that will replace <code>target</code>
   * @return The original string replaced with the proper strings
   */
  public static String replace(String main, String target, String replacement) {
    int i = main.indexOf(target);
    if (i == -1) {
      return main;
    }
    return main.substring(0, i) + replacement + main.substring(i + target.length());
  }
}
