package blackbox;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.regex.Pattern;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;


/**
 * Stores information about single FS gauge, including which variables it contains, which are
 * selected, sorting order, etc...
 *
 * @author Robbie McElrath
 */
public class Gauge extends JScrollPane implements ItemListener, ActionListener {
  private final static long serialVersionUID = 49923878324L;

  /** Stores the variables corresponding to this gauge */
  private VariableList vars;
  /** The File object corresponding to this gauge */
  private File loc;
  /** The list of the variables */
  private JPanel list;
  /** The tab Panel associated with this gauge */
  private JPanel tab;
  /** The text appearing on the tab */
  private JLabel title;
  /** The close button on the tab */
  private JButton close;

  /**
   * Constructs a new Gauge object from the specified File
   *
   * @param loc A File object corresponding to the location of the gauge
   */
  public Gauge(File loc) {
    this(loc, null);
  }

  /**
   * Constructs a new Gauge object from the specified File
   *
   * @param loc A File object corresponding to the location of the gauge
   * @param selected The list of variables to be pre-selected
   */
  public Gauge(File loc, ArrayList<String> selected) {
    super();
    setName(loc.getName());
    getVerticalScrollBar().setUnitIncrement(23);
    // create the tab
    tab = new JPanel(new BorderLayout(2, 0)); // use BorderLayout to get rid of padding
    tab.setOpaque(false);
    title = new JLabel(getName());
    close = new CloseButton();
    close.addActionListener(this);
    tab.add(title, BorderLayout.WEST);
    tab.add(close, BorderLayout.EAST);
    // parse the file & update display
    vars = new VariableList();
    this.loc = loc;
    reload(selected);
  }

  /**
   * Called when the Gauge should be closed
   */
  @Override
  public void actionPerformed(ActionEvent event) {
    if (event.getSource() == close) {
      BlackBox3 bb2 = getBlackBox2();
      if (bb2 != null) {
        bb2.close(this);
      }
    }
  }

  /**
   * Returns the File corresponding to this Gauge
   *
   * @return This gauge's source file
   */
  public File getFile() {
    return loc;
  }

  /**
   * Returns how the gauge is sorted
   *
   * @return A SortBy enumeration saying how it's currently sorted
   */
  public SortBy getSortedBy() {
    return vars.getSort();
  }

  /**
   * Checks if the given variable is present in the Gauge.
   *
   * @param var The variable to check for
   * @return <code>true</code> if the variable is in the Guage
   */
  public boolean contains(Variable var) {
    return vars.contains(var);
  }

  /**
   * Selects all the variables
   */
  public void selectAll() {
    for (Variable var : vars) {
      var.setSelected(true);
    }
  }

  /**
   * Deselects all the variables
   */
  public void deselectAll() {
    for (Variable var : vars) {
      var.setSelected(false);
    }
  }

  /**
   * Returns the tab for this Gauge
   *
   * @return The tab
   */
  public JPanel getTab() {
    return tab;
  }

  /**
   * Sorts the Gauge's variables with the given SortBy value
   *
   * @param sb What to sort by
   */
  public void sort(SortBy sb) {
    vars.sort(sb);
    redraw();
  }

  public enum SortBy {
    TYPE {
      @Override
      boolean ordered(Variable v1, Variable v2) {
        return v1.getType().toLowerCase().compareTo(v2.getType().toLowerCase()) < 0;
      }
    },
    NAME {
      @Override
      boolean ordered(Variable v1, Variable v2) {
        return v1.getName().toLowerCase().compareTo(v2.getName().toLowerCase()) < 0;
      }
    },
    UNITS {
      @Override
      boolean ordered(Variable v1, Variable v2) {
        return v1.getUnits().toLowerCase().compareTo(v2.getUnits().toLowerCase()) < 0;
      }
    },
    OCCURRENCE {
      @Override
      boolean ordered(Variable v1, Variable v2) {
        return v1.getIndex() < v2.getIndex();
      }
    },
    CHECKED {
      @Override
      boolean ordered(Variable v1, Variable v2) {
        return v1.isSelected() && !v2.isSelected();
      }
    };
    abstract boolean ordered(Variable v1, Variable v2);
  }

  /**
   * Reparses the file and updates the variable list to incorporate any changes made to it. If new
   * variables were found, they will be unselected. Existing variables will retain their state.
   */
  public void reload() {
    reload(null);
  }

  /**
   * Reparses the file and updates the variable list to incorporate any changes made to it. If new
   * variables were found, they will be unselected. Existing variables will retain their state.
   *
   * @param selected The list of variables to be pre-selected
   */
  public void reload(final ArrayList<String> selected) {
    new Thread(new Runnable() {
      @Override
      public void run() {
        final boolean saved = isSaved();
        // load variables
        VariableList vars2 = new VariableList();
        String w = "[ \t]*"; // whitespace
        Pattern regex =
            Pattern.compile(w + ">?" + w + "[AEPL]" + w + "\\:" + w + "[A-Za-z0-9_ \t,:]*" + w);
        try (BufferedReader in = new BufferedReader(new FileReader(loc))) {
          char[] cbuf = new char[127];
          int index = 0;
          int ix = 0;
          while (in.read(cbuf, index, 1) != -1) {
            switch (cbuf[index]) {
              case ')':
                String v = BlackBox3.replace(String.copyValueOf(cbuf, 0, index), "&gt;", ">");
                if (regex.matcher(v).matches()) {
                  vars2.add(new Variable(v, ix++));
                }
              case '(':
                index = 0;
                break;
              default:
                index++;
                if (index == cbuf.length) {
                  index = 0;
                }
                break;
            }
          }
        } catch (IOException e) {
        }
        // now restore state of existing variables
        VariableList sel = getSelectedList();
        for (Variable var : vars2) {
          if (selected != null) {
            var.setSelected(selected.contains(var.toString()));
          } else {
            var.setSelected(sel.contains(var));
          }
          var.addItemListener(Gauge.this);
        }
        vars = vars2;
        vars.setSort(SortBy.OCCURRENCE);
        SwingUtilities.invokeLater(new Runnable() {
          @Override
          public void run() {
            redraw();
            getBlackBox2().gaugeChanged();
            // set it as saved if it was
            if (saved) {
              setSaved();
            }
          }
        });
      }
    }).start();
  }

  /**
   * Called when a Variable has been checked or unchecked
   */
  @Override
  public void itemStateChanged(ItemEvent event) {
    BlackBox3 bb2 = getBlackBox2();
    if (bb2 != null) {
      bb2.setModified();
    }
    if (!title.getText().startsWith("*")) {
      title.setText("* " + title.getText());
    }
  }

  /**
   * Checks if the current gauge has been modified since the last save.
   *
   * @return true if the gauge's state has been modified (a variable checkbox has changed state)
   */
  public boolean isSaved() {
    return title.getText().charAt(0) != '*';
  }

  /**
   * Sets this gauge as saved.
   */
  public void setSaved() {
    if (title.getText().length() > 2 && title.getText().substring(0, 2).equals("* ")) {
      title.setText(title.getText().substring(2));
    }
  }

  /**
   * Gets an Iterator containing all the variables selected in this gauge.
   *
   * @return An Iterator of all selected Variables in this Gauge
   */
  public Iterator<Variable> getSelected() {
    return getSelectedList().iterator();
  }

  /**
   * Gets an ArrayList containing all the selected variables in this gauge
   *
   * @return An ArrayList of selected variables
   */
  private VariableList getSelectedList() {
    VariableList selected = new VariableList();
    for (Variable var : vars) {
      if (var.isSelected()) {
        selected.add(var);
      }
    }
    return selected;
  }

  /**
   * Returns whether the gauges has any selected variables or not
   *
   * @return true if one or more variables are selected
   */
  public boolean hasSelected() {
    for (Variable var : vars) {
      if (var.isSelected()) {
        return true;
      }
    }
    return false;
  }

  /**
   * Gets an Iterator containing all the Variables in this gauge.
   *
   * @return An Iterator of all variables in this Gauge, selected or not
   */
  public Iterator<Variable> getVariables() {
    return vars.iterator();
  }

  /**
   * Gets the BlackBox2 class that contains this Gauge object
   *
   * @return The containing BlackBox2 object
   */
  private BlackBox3 getBlackBox2() {
    Container el = this;
    while (el.getParent() != null && !(el instanceof BlackBox3)) {
      el = el.getParent();
    }
    return el != null ? ((BlackBox3) el) : null;
  }

  /**
   * Redraws the Gauge's panel (used if the order/contents of the gauge changes.
   */
  public void redraw() {
    list = new JPanel(new GridBagLayout());
    list.setBackground(Color.WHITE);
    GridBagConstraints c = BlackBox3.createConstraints();
    c.fill = GridBagConstraints.NONE;
    c.anchor = GridBagConstraints.WEST;
    for (Variable var : vars) {
      list.add(var, c);
      c.gridy++;
    }
    JLabel padding = new JLabel();
    c.weightx = 1.0;
    c.weighty = 1.0;
    list.add(padding, c);
    setViewportView(list);
  }

  /**
   * Checks if two Gauges are the same (same source file)
   */
  @Override
  public boolean equals(Object o) {
    if (o instanceof Gauge) {
      return ((Gauge) o).getFile().equals(getFile());
    }
    return false;
  }

  /**
   * Custom close button
   */
  private class CloseButton extends JButton {
    private final static long serialVersionUID = 50987234L;

    public CloseButton() {
      setIcon(BlackBox3.getIcon("close"));
      setPreferredSize(new Dimension(16, 16));
      setContentAreaFilled(false);
      setOpaque(false);
    }
  }
}
