package blackbox;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Scanner;

import javax.swing.AbstractAction;
import javax.swing.JOptionPane;

public class Output {
  /** Character width of the generated BlackBox gauge */
  private final static int CHAR_WIDTH = 6;
  /** Character height of the generated BlackBox gauge */
  private final static int CHAR_HEIGHT = 7;
  /** Grid width of the generated BlackBox gauge */
  private final static int GRID_WIDTH = CHAR_WIDTH;
  /** Grid height of the generated BlackBox gauge */
  private final static int GRID_HEIGHT = CHAR_HEIGHT + 1;
  /** Scrollbar Size */
  private final static int SCROLLBAR_SIZE = GRID_HEIGHT;

  /** The height of the generated gauge */
  int height = 500;
  /** The width of the generated gauge */
  int width = 500;
  /** Background color */
  String backgroundColor = "Black";
  /** Variable color */
  String variableColor = "White";
  /** Title color */
  String titleColor = "#00ff00";
  /** Unselected tab text color */
  String tabColor = "#gray";
  /** Unselected tab background color */
  String tabBackground = "080808";
  /** Selected tab text color */
  String tabSelectedColor = "#008800";
  /** Selected tab background color */
  String tabSelectedBackground = "#dddddd";
  /** Font */
  String font = "Arial";
  /** Path to logger file directory */
  String loggerPath = "";
  /** OutputFormat to use */
  Format format;
  /** The default display format used during output */
  Mapping mappingDefault;
  /** Name mapping. Maps variable name to a certain display format. */
  ArrayList<Mapping> mappingName;
  /** Units mapping. Maps variable units to display format for that unit. */
  ArrayList<Mapping> mappingUnits;

  /** BlackBox3 object */
  private BlackBox3 bb;
  /** How many rows the tabs take up */
  private int tabHeight = 0;

  /**
   * Constructor for the Output object
   *
   * @param bb3 BlackBox2 instance
   */
  Output(BlackBox3 bb3) {
    this();
    bb = bb3;
  }

  /**
   * Default constructor
   */
  Output() {
    format = column;
  }

  /**
   * Sets the BlackBox2 Object associated with this output
   */
  void setBlackBox3(BlackBox3 bb3) {
    bb = bb3;
  }

  // ---------------------------------------------------------------------------------- OUTPUT VIEWS
  /**
   * Multi-tabbed output format
   */
  Format multiTabbed = new Format(
      "Tabbed Multiple Output",
      "Tabbed Output with multiple gauges open at a time") {
    private final static long serialVersionUID = 87932809234325L;

    @Override
    public String getName() {
      return "Multi-Tabbed";
    }

    @Override
    Dimension printVars(PrintStream header, PrintStream body, PrintStream footer,
        PrintStream mouse, BlackBox3.GaugeIterator gauges) {
      Dimension d = new Dimension(0, 0);
      Gauge gauge;
      Rectangle[] tabs = printTabs(footer, gauges);
      int y, xMax;
      header.println("\t<Macro Name=\"BB2Shift\">");
      header.print("\t\t0 ");
      gauges.reset();
      while (gauges.hasNext()) {
        gauge = gauges.next();
        String name = getGaugeName(gauge);
        int i = gauges.getIndex();
        Rectangle tab = tabs[i];
        body.println("\t\t<Element>");
        body.println("\t\t\t<Visible>(L:BB2Gauge" + i + "Vis,number)</Visible>");
        // mouse text
        mouse.println("\t\t<Area Left=\"" + (tab.left * GRID_WIDTH) + "\" Right=\""
            + (tab.right * GRID_WIDTH) + "\" Top=\"" + (tab.top * GRID_HEIGHT) + "\" Bottom=\""
            + (tab.bottom * GRID_HEIGHT) + "\">");
        mouse.println("\t\t\t<Cursor Type=\"Hand\" />");
        mouse.print("\t\t\t<Click Kind=\"LeftSingle\">");
        mouse.print("(L:BB2Gauge" + i + "Vis,number) ! (&gt;L:BB2Gauge" + i
            + "Vis,number) @BB2Shift");
        mouse.println("</Click>");
        mouse.println("\t\t</Area>");
        // tab contents
        body.println("\t\t<Element>");
        y = tabs[tabs.length - 1].bottom;
        xMax = printString(body, name, mappingDefault.getFormatWidth() + 2, y++);
        Iterator<Variable> vars = gauge.getSelected();
        while (vars.hasNext()) {
          xMax = Math.max(xMax, printVariable(body, vars.next(), 1, y++));
        }
        body.println("\t\t\t<Shift>");
        body.println("\t\t\t\t<Value Minimum=\"0\" Maximum=\"" + d.width * GRID_WIDTH
            + "\">(L:BB2Gauge" + i + "Shift,number)</Value>");
        body.println("\t\t\t\t<Scale X=\"1\" />");
        body.println("\t\t\t</Shift>");
        body.println("\t\t</Element>");
        body.println("\t\t</Element>");
        header.print("d (&gt;L:BB2Gauge" + i + "Shift,number) " + xMax * GRID_WIDTH
            + " (L:BB2Gauge" + i + "Vis,number) * + ");
        d.width += xMax;
        d.height = Math.max(d.height, y);
      }
      header.println("p");
      header.println("\t</Macro>");
      d.height *= GRID_HEIGHT;
      d.width *= GRID_WIDTH;
      return d;
    }
  };

  /**
   * Tabbed output format
   */
  Format tabbed = new Format(
      "Tabbed Single Output",
      "Tabbed Output with each gauge getting its own tab") {
    private final static long serialVersionUID = 38798237498123L;
    private Dimension d;
    private int x, y, xMax, yMin;

    @Override
    public String getName() {
      return "Tabbed";
    }

    /**
     * Outputs the gauge data
     */
    @Override
    Dimension printVars(PrintStream header, PrintStream body, PrintStream footer,
        PrintStream mouse, BlackBox3.GaugeIterator gauges) {
      d = new Dimension(0, (height - SCROLLBAR_SIZE) / GRID_HEIGHT - 1);
      Gauge gauge;
      Rectangle[] tabs = printTabs(footer, gauges);
      gauges.reset();
      while (gauges.hasNext()) {
        gauge = gauges.next();
        Rectangle tab = tabs[gauges.getIndex()];
        body.println("\t\t<Element>");
        body.println("\t\t\t<Visible>(L:BB2Gauge" + gauges.getIndex() + "Vis,number)</Visible>");
        // mouse text
        mouse.println("\t\t<Area Left=\"" + (tab.left * GRID_WIDTH) + "\" Right=\""
            + (tab.right * GRID_WIDTH) + "\" Top=\"" + (tab.top * GRID_HEIGHT) + "\" Bottom=\""
            + (tab.bottom * GRID_HEIGHT) + "\">");
        mouse.println("\t\t\t<Cursor Type=\"Hand\" />");
        mouse.print("\t\t\t<Click Kind=\"LeftSingle\">");
        for (int i = 0; i < gauges.getSize(); i++) {
          mouse.print((i == gauges.getIndex() ? 1 : 0) + " (&gt;L:BB2Gauge" + i + "Vis,number) ");
        }
        mouse.println("</Click>");
        mouse.println("\t\t</Area>");
        // tab contents
        x = xMax = 1;
        y = yMin = tabs[tabs.length - 1].bottom;
        Iterator<Variable> vars = gauge.getSelected();
        while (vars.hasNext()) {
          xMax = Math.max(xMax, printVariable(body, vars.next(), x, y++));
          checkCoords();
        }
        d.width = Math.max(d.width, xMax);
        body.println("\t\t</Element>");
      }
      d.height *= GRID_HEIGHT;
      d.width *= GRID_WIDTH;
      return d;
    }

    private void checkCoords() {
      if (y > d.height) {
        x = xMax;
        y = yMin;
      }
    }
  };

  /**
   * Single column output format
   */
  Format column = new Format("Single Column Output", "Output in a single column") {
    private final static long serialVersionUID = 387982374987123L;

    @Override
    public String getName() {
      return "Column";
    }

    /**
     * Outputs the gauge data
     */
    @Override
    Dimension printVars(PrintStream header, PrintStream body, PrintStream footer,
        PrintStream mouse, BlackBox3.GaugeIterator gauges) {
      Dimension d = new Dimension(0, 1);
      while (gauges.hasNext()) {
        Gauge gauge = gauges.next();
        d.width =
            Math.max(
                d.width,
                printString(body, getGaugeName(gauge), mappingDefault.getFormatWidth() + 2,
                    d.height++));
        Iterator<Variable> vars = gauge.getSelected();
        while (vars.hasNext()) {
          d.width = Math.max(d.width, printVariable(body, vars.next(), 1, d.height++));
        }
      }
      d.height = ++d.height * GRID_HEIGHT;
      d.width *= GRID_WIDTH;
      return d;
    }
  };

  /**
   * Single column wrapped output format
   */
  Format wrapped = new Format("Wrapped Output", "Outputs vertically - wraps to multiple columns") {
    private final static long serialVersionUID = 68387374987123L;
    private Dimension d;
    private int x, y, yMax, xMax;
    private Gauge gauge;

    @Override
    public String getName() {
      return "Wrapped";
    }

    /**
     * Outputs the gauge data
     */
    @Override
    Dimension printVars(PrintStream header, PrintStream body, PrintStream footer,
        PrintStream mouse, BlackBox3.GaugeIterator gauges) {
      d = new Dimension();
      x = y = 1;
      yMax = (height - SCROLLBAR_SIZE) / GRID_HEIGHT - 1;
      xMax = 0;
      while (gauges.hasNext()) {
        gauge = gauges.next();
        xMax = Math.max(
            xMax,
            printString(body, getGaugeName(gauge), x + mappingDefault.getFormatWidth() + 1, y++));
        checkCoords(body);
        Iterator<Variable> vars = gauge.getSelected();
        while (vars.hasNext()) {
          xMax = Math.max(xMax, printVariable(body, vars.next(), x, y++));
          checkCoords(body);
        }
      }
      d.width = xMax * GRID_WIDTH;
      d.height = yMax * GRID_HEIGHT;
      return d;
    }

    private void checkCoords(PrintStream out) {
      if (y > yMax) {
        x = xMax;
        y = 1;
        xMax = printString(out, getGaugeName(gauge), x + mappingDefault.getFormatWidth() + 1, y++);
      }
    }
  };

  /**
   * Logger CSV output format
   */
  Format loggerCsv = new Format("Logger - CSV", "Outputs using Logger in CSV format") {
    private static final long serialVersionUID = -87239747125247L;

    @Override
    String getName() {
      return "LoggerCsv";
    }

    @Override
    protected String getTemplate() {
      return "res/loggerTemplate.xml";
    }

    /**
     * Outputs the gauge data
     */
    @Override
    Dimension printVars(PrintStream header, PrintStream body, PrintStream footer,
        PrintStream mouse, BlackBox3.GaugeIterator gauges) {
      header.println("<Macro Name=\"Path\">'" + loggerPath + "'</Macro>");
      // Code for opening the file
      footer.println("(>C:LOGGER:openCsvWrite)");
      while (gauges.hasNext()) {
        Gauge gauge = gauges.next();
        Iterator<Variable> vars = gauge.getSelected();
        while (vars.hasNext()) {
          Variable v = vars.next();
          String title = "'" + v.getType() + ":" + v.getName() + "," + v.getUnits() + "'";
          footer.println(title + " (>C:LOGGER:string)");
        }
      }
      footer.println("1 (>C:LOGGER:newline)");
      // Dumping the variables
      gauges = gauges.reset();
      while (gauges.hasNext()) {
        Gauge gauge = gauges.next();
        Iterator<Variable> vars = gauge.getSelected();
        while (vars.hasNext()) {
          Variable v = vars.next();
          if (v.getUnits().equalsIgnoreCase("string")) {
            body.println(v.toString() + " (>C:LOGGER:string)");
          } else {
            body.println(v.toString() + " (>C:LOGGER:number)");
          }
        }
      }
      body.println("1 (>C:LOGGER:newline)");
      return new Dimension(200, 200);
    }
  };

  /**
   * Logger line output format
   */
  Format loggerLine =
      new Format("Logger - Newline", "Outputs using Logger in standard line format") {
        private static final long serialVersionUID = -87239747442358L;

        @Override
        String getName() {
          return "LoggerLine";
        }

        @Override
        protected String getTemplate() {
          return "res/loggerTemplate.xml";
        }

        /**
         * Outputs the gauge data
         */
        @Override
        Dimension printVars(PrintStream header, PrintStream body, PrintStream footer,
            PrintStream mouse, BlackBox3.GaugeIterator gauges) {
          header.println("<Macro Name=\"Path\">'" + loggerPath + "'</Macro>");
          footer.println("(>C:LOGGER:openWrite)");
          while (gauges.hasNext()) {
            Gauge gauge = gauges.next();
            Iterator<Variable> vars = gauge.getSelected();
            while (vars.hasNext()) {
              Variable v = vars.next();
              if (v.getUnits().equalsIgnoreCase("string")) {
                body.println(v.toString() + " (>C:LOGGER:string)");
              } else {
                body.println(v.toString() + " (>C:LOGGER:number)");
              }
            }
          }
          return new Dimension(200, 200);
        }
      };

  /**
   * Action template for output format radio items
   */
  abstract class Format extends AbstractAction {
    private static final long serialVersionUID = 3790708370245681852L;

    /**
     * Creates a new Action with the specified properties. Leave any of the parameters null if you
     * don't wish to use them
     *
     * @param text The name of the action (will appear on a button or menu item)
     * @param desc A short description of the event for use in a tooltip
     */
    public Format(String name, String desc) {
      super(name);
      putValue(AbstractAction.SHORT_DESCRIPTION, desc);
    }

    /**
     * Called when the user selects the format from the menu
     */
    @Override
    public void actionPerformed(ActionEvent event) {
      if (format != this) {
        bb.setModified();
        format = this;
      }
    }

    /**
     * Generates the main body of the BlackBox gauge
     */
    public void print(PrintStream out, BlackBox3.GaugeIterator gauges) {
      ByteArrayOutputStream outHeader = new ByteArrayOutputStream(2000);
      ByteArrayOutputStream outBody = new ByteArrayOutputStream(10000);
      ByteArrayOutputStream outFooter = new ByteArrayOutputStream(2000);
      ByteArrayOutputStream outMouse = new ByteArrayOutputStream(2000);
      tabHeight = 0;
      Dimension d = printVars(new PrintStream(outHeader), new PrintStream(outBody),
          new PrintStream(outFooter), new PrintStream(outMouse), gauges);
      int[] dimensions =
          {SCROLLBAR_SIZE, height, width, d.height, d.width, height - SCROLLBAR_SIZE,
              width - SCROLLBAR_SIZE, height - SCROLLBAR_SIZE - 72, width - SCROLLBAR_SIZE - 72,
              d.height - height + SCROLLBAR_SIZE, d.width - width + SCROLLBAR_SIZE,
              SCROLLBAR_SIZE / 2, CHAR_HEIGHT, CHAR_WIDTH, GRID_HEIGHT, GRID_WIDTH,
              tabHeight * GRID_HEIGHT};
      Replacer r = new Replacer(dimensions);
      try {
        Scanner in = new Scanner(BlackBox3.class.getResourceAsStream(getTemplate()));
        String line;
        while (in.hasNextLine()) {
          line = in.nextLine();
          if (line.trim().equals("$header$")) {
            if (backgroundColor.toLowerCase().equals("black") || backgroundColor.equals("#000000")) {
              out.println("\t<Size X=\"" + width + "\" Y=\"" + height + "\" />");
            } else {
              out.println("\t<Element>");
              out.println("\t\t<Position X=\"0\" Y=\"0\" />");
              out.println("\t\t<Rectangle Width=\"" + width + "\" Height=\"" + height
                  + "\" FillColor=\"" + backgroundColor + "\" Bright=\"Yes\" />");
              out.println("\t</Element>");
            }
            outHeader.writeTo(out);
          } else if (line.trim().equals("$body$")) {
            outBody.writeTo(out);
          } else if (line.trim().equals("$footer$")) {
            outFooter.writeTo(out);
          } else if (line.trim().equals("$mouse$")) {
            outMouse.writeTo(out);
          } else {
            out.println(r.replace(line));
          }
        }
        in.close();
      } catch (IOException e) {
        JOptionPane.showMessageDialog(bb, "Error writing gauge!", "Error!",
            JOptionPane.ERROR_MESSAGE);
      }
    }

    /**
     * Returns the path of the template that this output format should use
     *
     * @return The path
     */
    protected String getTemplate() {
      return "res/template.xml";
    }

    /**
     * Returns the name of this format
     *
     * @return This OutputFormat's name
     */
    abstract String getName();

    /**
     * Outputs the gauge to the given PringStream
     *
     * @return The Dimensions in pixels of the generated gauge
     */
    abstract Dimension printVars(PrintStream header, PrintStream body, PrintStream footer,
        PrintStream mouse, BlackBox3.GaugeIterator gauges);
  }

  // -------------------------------------------------------------------------------- HELPER METHODS
  /**
   * Prints out the tabs on the top of the gauge and returns an array of their locations
   *
   * @param out The PrintStream the tabs should be printed to
   * @param gauges The GaugeIterator containing the guages
   * @return An array with all the tab locations
   */
  private Rectangle[] printTabs(PrintStream out, BlackBox3.GaugeIterator gauges) {
    int x = 0, y = 0;
    Rectangle[] rects = new Rectangle[gauges.getSize()]; // stores locations of tabs for click-spots
    printRect(out, 0);
    while (gauges.hasNext()) {
      Gauge gauge = gauges.next();
      String name = getGaugeName(gauge);
      if (x + name.length() > width / GRID_WIDTH && x > 0) {
        x = 0;
        printRect(out, ++y);
      }
      rects[gauges.getIndex()] = new Rectangle(x, y, 1, x + name.length() + 2);
      printString(out, " " + name + " ", x, y, tabSelectedColor, tabSelectedBackground);
      out.println("\t\t<Element>");
      out.println("\t\t\t<Visible>(L:BB2Gauge" + gauges.getIndex() + "Vis,number) !</Visible>");
      x = printString(out, " " + name + " ", x, y, tabColor, tabBackground) + 1;
      out.println("\t\t</Element>");
    }
    tabHeight = y + 1;
    return rects;
  }

  /**
   * Outputs a rectangle at the given y grid location that takes up an entire row
   *
   * @param y The row to place the rectangle on
   */
  private void printRect(PrintStream out, int y) {
    out.println("\t<Element>");
    out.println("\t\t<Position X=\"0\" Y=\"" + (y * GRID_HEIGHT) + "\" />");
    out.println("\t\t<Rectangle Width=\"" + width + "\" Height=\"" + GRID_HEIGHT
        + "\" FillColor=\"" + backgroundColor + "\" Bright=\"Yes\" />");
    out.println("\t</Element>");
  }

  /**
   * Prints an FSXML Element tag containing the given string at the given location. The color is the
   * default title color and the background color is null.
   *
   * @param out The PrintStream to output to
   * @param str The String to display
   * @param x The X position the text should appear in (relative to the grid)
   * @param y The Y position the text should appear in (relative to the grid)
   * @return The new x position
   */
  private int printString(PrintStream out, String str, int x, int y) {
    return printString(out, str, x, y, titleColor, null);
  }

  /**
   * Prints an FSXML Element tag containing the given string at the given location
   *
   * @param out The PrintStream to output to
   * @param str The String to display
   * @param x The X position the text should appear in (relative to the grid)
   * @param y The Y position the text should appear in (relative to the grid)
   * @param color The font color
   * @param hl The highlight color
   * @return The new x position
   */
  private int printString(PrintStream out, String str, int x, int y, String color, String hl) {
    out.println("\t\t\t<Element>");
    out.println("\t\t\t\t<Position X=\"" + (x * GRID_WIDTH) + "\" Y=\"" + (y * GRID_HEIGHT)
        + "\" />");
    out.println("\t\t\t\t<Text X=\"" + (str.length() * GRID_WIDTH) + "\" Y=\"" + CHAR_HEIGHT
        + "\" " + "Bright=\"Yes\" Length=\"" + str.length() + "\" Font=\"" + font + "\" Color=\""
        + color + "\" " + "Adjust=\"Center\" " + "VerticalAdjust=\"Center\" "
        + "Multiline=\"No\" Fixed=\"Yes\" " + (hl == null ? "" : "BackgroundColor=\"" + hl + "\" ")
        + ">");
    out.println("\t\t\t\t\t<String>" + str + "</String>");
    out.println("\t\t\t\t</Text>");
    out.println("\t\t\t</Element>");
    return x + str.length();
  }

  /**
   * Prints an FSXML Element tag containing the given Variable at the given location. It is
   * formatted according to the format mappings.
   *
   * @param out The PrintStream to output to
   * @param str The Variable to display
   * @param x The X position the text should appear in (relative to the grid)
   * @param y The Y position the text should appear in (relative to the grid)
   */
  private int printVariable(PrintStream out, Variable var, int x, int y) {
    Mapping map = mappingDefault;
    int index;
    if ((index = mappingUnits.indexOf(new Mapping(var.getUnits(), null))) >= 0) {
      map = mappingUnits.get(index);
    }
    if ((index = mappingName.indexOf(new Mapping(var.getName(), null))) >= 0) {
      map = mappingName.get(index);
    }
    int nameLength = var.getName().length() + 3;
    if (bb.showUnitsCheckBox.isSelected()) {
      nameLength += var.getUnits().length() + 2;
    }
    int length = nameLength + map.getFormatWidth();
    String text = "%(" + var + ")%!" + map.getFormat() + "!%(' " + var.getType() + ":"
        + var.getName() + (bb.showUnitsCheckBox.isSelected() ? ", " + var.getUnits() : "")
        + "')%!-" + nameLength + "s!";
    out.println("\t\t\t<Element>");
    out.println("\t\t\t\t<Position X=\"" + (x * GRID_WIDTH) + "\" Y=\"" + (y * GRID_HEIGHT)
        + "\" />");
    out.println("\t\t\t\t<Text X=\"" + (length * GRID_WIDTH) + "\" Y=\"" + CHAR_HEIGHT
        + "\" Bright=\"" + "Yes\" Length=\"" + length + "\" Font=\"" + font + "\" Color=\""
        + variableColor + "\" Adjust="
        + "\"Center\" VerticalAdjust=\"Center\" Multiline=\"No\" Fixed=\"Yes\">");
    out.println("\t\t\t\t\t<String>" + text + "</String>");
    out.println("\t\t\t\t</Text>");
    out.println("\t\t\t</Element>");
    return x + length;
  }

  /**
   * Returns a Gauge's name without the xml extension
   */
  private String getGaugeName(Gauge g) {
    String name = g.getFile().getName();
    return name.toLowerCase().endsWith(".xml") ? name.substring(0, name.length() - 4) : name;
  }


  // ------------------------------------------------------------------------------- RECTANGLE CLASS
  private class Rectangle {
    int top, left, bottom, right;

    public Rectangle(int x, int y, int height, int width) {
      this.left = x;
      this.top = y;
      this.bottom = y + height;
      this.right = x + width;
    }
  }


  // -------------------------------------------------------------------------------- REPLACER CLASS
  private class Replacer {
    private String[] keys, values;

    public Replacer(int[] dimensions) {
      keys = new String[dimensions.length];
      values = new String[dimensions.length];
      for (int i = 0; i < dimensions.length; i++) {
        keys[i] = "\\$" + i + "\\$";
        values[i] = "" + (dimensions[i] < 0 ? 0 : dimensions[i]);
      }
    }

    public String replace(String str) {
      if (str.matches(".*\\$[0-9]{1,2}\\$.*")) {
        for (int i = 0; i < keys.length; i++) {
          str = str.replaceAll(keys[i], values[i]);
        }
      }
      return str;
    }
  }


  // --------------------------------------------------------------------------------- MAPPING CLASS
  /**
   * This class stores the mapping information used to relate variable units or name to a certain
   * formatting. This allows the BlackBox output gauge to be customized to display certain variables
   * in certain ways.
   */
  static final class Mapping {
    private String key, format;

    public Mapping(String key, String format) {
      this.key = key;
      this.format = format;
    }

    public String getKey() {
      return key;
    }

    public String getFormat() {
      return format;
    }

    public int getFormatWidth() {
      int i = 0, num = 0;
      while (i < format.length() && format.charAt(i) >= '0' && format.charAt(i) <= '9') {
        num *= 10;
        num += (format.charAt(i) - '0');
        i++;
      }
      return num;
    }

    @Override
    public boolean equals(Object o) {
      if (o instanceof Mapping) {
        return ((Mapping) o).getKey().toLowerCase().equals(getKey().toLowerCase());
      }
      return false;
    }
  }
}
