package blackbox;

import java.awt.Font;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.JCheckBox;


/**
 * Represents a single FS variable
 *
 * @author Robbie McElrath
 */
public class Variable extends JCheckBox implements ItemListener {
  private final static long serialVersionUID = -987832423L;

  /** Font used when variable is unselected */
  private final static Font unselected = new Font("SansSerif", Font.PLAIN, 11);
  /** Font used when variable is selected */
  private final static Font selected = new Font("SansSerif", Font.BOLD, 11);

  /** Variable's type */
  private String type;
  /** Variable's name */
  private String name;
  /** Variable's units */
  private String units;
  /** Variable's initial index */
  private int index;

  /**
   * Constructs a new Variable from the given signature. The signature should be the FS9 XML
   * variable referencing format: (T:name,units) It is assumed that the signature is in the correct
   * form. The index will be set to -1
   *
   * @param signature The signature
   */
  public Variable(String signature) {
    this(signature, -1);
  }

  /**
   * Constructs a new Variable from the given signature and the given index. The signature should be
   * the FS9 XML variable referencing format: (T:name,units) OR (>T:name,units) It is assumed that
   * the signature is in the correct form.
   *
   * @param signature The signature
   * @param idx The index to assign the Variable to
   */
  public Variable(String signature, int idx) {
    int s = signature.indexOf('>');
    int m = signature.indexOf(':');
    int e, e2 = e = signature.indexOf(',');
    s = (s == -1 ? 0 : s + 1);
    e = (e == -1 ? signature.length() : e);
    type = signature.substring(s, m).trim();
    name = signature.substring(m + 1, e).trim();
    units = e2 == -1 ? "" : signature.substring(e + 1, signature.length()).trim();

    index = idx;
    setSelected(false);
    setText(toString());
    setFont(unselected);
    setOpaque(false);
    addItemListener(this);
  }

  /**
   * Called when the box is checked/unchecked
   */
  @Override
  public void itemStateChanged(ItemEvent event) {
    setFont(isSelected() ? selected : unselected);
  }

  /**
   * Returns the FS9 variable name
   *
   * @return A String representing this object
   */
  @Override
  public String toString() {
    if (units.equals("")) {
      return "(" + type + ":" + name + ")";
    }
    return "(" + type + ":" + name + "," + units + ")";
  }

  /**
   * Gets the variable's type
   */
  public String getType() {
    return type;
  }

  /**
   * Gets the variable's name
   */
  @Override
  public String getName() {
    return name;
  }

  /**
   * Gets the variable's units
   */
  public String getUnits() {
    return units;
  }

  /**
   * Gets the variable's index
   */
  public int getIndex() {
    return index;
  }

  /**
   * Compares this Variable with another
   *
   * @param o The object to compare it to
   * @return <code>true</code> if the objects are the same
   */
  @Override
  public boolean equals(Object o) {
    if (o instanceof Variable) {
      return toString().toLowerCase().equals(o.toString().toLowerCase());
    }
    return false;
  }

  /**
   * Returns the has code value for this object
   */
  @Override
  public int hashCode() {
    return toString().hashCode();
  }
}
