package blackbox;

import java.util.HashSet;
import java.util.Iterator;


/**
 * Stores a list of variables in sorted order
 *
 * @author Robbie McElrath
 */
public class VariableList implements Iterable<Variable> {
  /** The head of the VariableList */
  private Entry head;
  /**
   * The Tail of the VariableList, used for quickly inserting in occurrence mode. Note that this
   * variable isn't the correct value for any sorting mode other than occurrence.
   */
  private Entry tail;
  /** Hashtable of all Variables currently in this list */
  private HashSet<Variable> vars;
  /** Sorting method */
  private Gauge.SortBy sort;

  /**
   * Initializes this VariableList
   */
  public VariableList() {
    vars = new HashSet<Variable>(100);
  }

  /**
   * Checks if this VariableList contains a certain Variable
   */
  public boolean contains(Variable v) {
    return vars.contains(v);
  }

  /**
   * Adds the given variable to the list if it isn't already in it
   *
   * @param v The Variable to add
   */
  public void add(Variable v) {
    if (!vars.contains(v)) {
      head = add(v, head);
    }
  }

  /**
   * Adds the given variable to the specified list if it isn't already added. Note that it is
   * assumed the variable isn't already in the gauge
   *
   * @param v The Variable to add
   * @param list The list to add the Variable to
   * @return The head element of the new list
   */
  private Entry add(Variable v, Entry list) {
    if (list == null) {
      return tail = new Entry(v, null);
    }
    // the list already has something in it
    if (sort == null) {
      tail.next = new Entry(v, null);
      tail = tail.next;
      return list;
    }
    // they're not sorting by occurrence
    if (sort.ordered(v, list.var)) {
      return new Entry(v, list);
    }
    // it's not the first element
    Entry e = list;
    while (e.next != null && sort.ordered(e.next.var, v)) {
      e = e.next;
    }
    e.next = new Entry(v, e.next);
    return list;
  }

  /**
   * Returns the SortBy used to sort this VariableList
   *
   * @return The SortBy used
   */
  public Gauge.SortBy getSort() {
    return sort;
  }

  /**
   * Sorts this VariableList with the given comparator
   *
   * @param sort The SoryBy to use while sorting
   */
  public void sort(Gauge.SortBy sort) {
    setSort(sort);
    Entry newHead = null;
    for (Variable var : this) {
      newHead = add(var, newHead);
    }
    head = newHead;
  }

  /**
   * Sets the sorting to be used in thie VariableList
   *
   * @param sort The sort to use
   */
  void setSort(Gauge.SortBy sort) {
    this.sort = sort;
  }

  /**
   * Returns an iterator over this list of Variables
   */
  @Override
  public Iterator<Variable> iterator() {
    return new VarIterator();
  }

  /**
   * Variable iterator
   */
  private class VarIterator implements Iterator<Variable> {
    private Entry e;

    private VarIterator() {
      e = head;
    }

    @Override
    public boolean hasNext() {
      return e != null;
    }

    @Override
    public Variable next() {
      Variable v = e.var;
      e = e.next;
      return v;
    }

    @Override
    public void remove() {}
  }

  /**
   * An entry in the list
   */
  private class Entry {
    Variable var;
    Entry next;

    private Entry(Variable var, Entry next) {
      this.var = var;
      this.next = next;
      vars.add(var);
    }

    @Override
    public String toString() {
      return var.toString() + "->" + (next == null ? "" : next.var.toString());
    }
  }

  /**
   * Returns a String representation of this VariableList
   */
  @Override
  public String toString() {
    String s = "";
    for (Variable v : this) {
      s += v.toString() + "->";
    }
    return s;
  }
}
